import copy
import datetime
import uuid
from dataclasses import dataclass, field, fields, is_dataclass
from decimal import Decimal
from enum import Enum, unique
from typing import Any, Callable, Dict, Tuple, Type, TypeVar, Optional, Union

from dacite import Config as DaciteConfig, from_dict, exceptions as daciteExceptions
from dacite.types import is_union
from dateutil.parser import parse

from mqer import pjson
from mqer.enum_mixin import is_enum
from mqer.types import EmptyObj
from mqer.utils import force_str

TDataclass = TypeVar("TDataclass", bound="BaseDataclassMixin")
Config = DaciteConfig


class ConvertDataclassError(Exception):
    pass


def datetime_parse(datetime_value: Any) -> Union[Tuple[datetime.datetime, Any], datetime.datetime]:
    if isinstance(datetime_value, datetime.datetime):
        return datetime_value
    return parse(datetime_value, fuzzy_with_tokens=False)


class WrongTypeError(daciteExceptions.DaciteFieldError):
    def __init__(self, dataclass_cls: Type[TDataclass],  field_type: Type, value: Any, field_path: Optional[str] = None) -> None:
        super().__init__(field_path=field_path)
        self.field_type = field_type
        self.value = value
        self.dataclass_cls = dataclass_cls

    def __str__(self) -> str:
        return (
            f'wrong value type for field "{self.field_path}" of dataclass model "{self.dataclass_cls}" - should be "{daciteExceptions._name(self.field_type)}" '
            f'instead of value "{self.value}" of type "{daciteExceptions._name(type(self.value))}"'
        )


@unique
class DataclassAsdictModeEnum(Enum):
    # use dataclass __dataclass_fields__ attribute
    DEFINE_FIELDS = 1
    # use __dict__ attribute
    DICT_ATTRIBUTE = 2


def _convert_dict_encoder(obj: Any) -> Tuple[Any, bool]:
    if hasattr(obj, "todict") and callable(getattr(obj, "todict", None)):
        return obj.todict(), True
    elif hasattr(obj, "tojson") and callable(getattr(obj, "tojson", None)):
        return obj.tojson(), True
    elif is_enum(obj):
        return getattr(obj, "value", force_str(obj)), True
    # For Date Time string spec, see ECMA 262
    # https://ecma-international.org/ecma-262/5.1/#sec-15.9.1.15
    # elif isinstance(obj, Promise):
    #     return force_str(obj), True
    elif isinstance(obj, datetime.datetime):
        representation = obj.isoformat()
        if representation.endswith('+00:00'):
            representation = representation[:-6] + 'Z'
        return representation, True
    elif isinstance(obj, datetime.date):
        return obj.isoformat(), True
    elif isinstance(obj, datetime.time):
        representation = obj.isoformat()
        return representation, True
    elif isinstance(obj, datetime.timedelta):
        return str(obj.total_seconds()), True
    elif isinstance(obj, Decimal):
        # Serializers will coerce decimals to strings by default.
        return float(obj), True
    elif isinstance(obj, uuid.UUID):
        return str(obj), True
    elif isinstance(obj, bytes):
        # Best-effort for binary blobs. See #4187.
        return obj.decode(), True
    elif hasattr(obj, 'tolist'):
        # Numpy arrays and array scalars.
        return obj.tolist(), True

    return obj, False


def _asdict_inner(obj: Any, dict_factory: Callable = dict, mode: DataclassAsdictModeEnum = DataclassAsdictModeEnum.DICT_ATTRIBUTE) -> Any:
    if is_dataclass(obj):
        result = []
        if mode == DataclassAsdictModeEnum.DICT_ATTRIBUTE:
            for k, v in obj.__dict__.items():
                if isinstance(v, EmptyObj):
                    continue
                value, has_convert = _convert_dict_encoder(v)
                if not has_convert:
                    value = _asdict_inner(v, dict_factory)
                result.append((k, value))
        else:
            for f in fields(obj):
                # Can be has field(repr=False, init=False)
                if not hasattr(obj, f.name):
                    continue
                v = getattr(obj, f.name)
                if isinstance(v, EmptyObj):
                    continue
                value, has_convert = _convert_dict_encoder(v)
                if not has_convert:
                    value = _asdict_inner(v, dict_factory)
                result.append((f.name, value))
        return dict_factory(result)
    elif isinstance(obj, tuple) and hasattr(obj, '_fields'):
        # obj is a namedtuple.  Recurse into it, but the returned
        # object is another namedtuple of the same type.  This is
        # similar to how other list- or tuple-derived classes are
        # treated (see below), but we just need to create them
        # differently because a namedtuple's __init__ needs to be
        # called differently (see bpo-34363).

        # I'm not using namedtuple's _asdict()
        # method, because:
        # - it does not recurse in to the namedtuple fields and
        #   convert them to dicts (using dict_factory).
        # - I don't actually want to return a dict here.  The main
        #   use case here is json.dumps, and it handles converting
        #   namedtuples to lists.  Admittedly we're losing some
        #   information here when we produce a json list instead of a
        #   dict.  Note that if we returned dicts here instead of
        #   namedtuples, we could no longer call asdict() on a data
        #   structure where a namedtuple was used as a dict key.

        return type(obj)(*[_asdict_inner(v, dict_factory) for v in obj])
    elif isinstance(obj, (list, tuple)):
        # Assume we can create an object of this type by passing in a
        # generator (which is not true for namedtuples, handled
        # above).
        return type(obj)(_asdict_inner(v, dict_factory) for v in obj)
    elif isinstance(obj, dict):
        return type(obj)((_asdict_inner(k, dict_factory), _asdict_inner(v, dict_factory)) for k, v in obj.items())
    else:
        return copy.deepcopy(obj)


class BaseDataclassMixin:
    @classmethod
    def _convert_union_field_type(cls, field_type: Any, field_value: Any) -> Any:
        is_optional = type(None) in field_type.__args__
        if is_optional and field_value is None:
            return None
        for union_item_type in field_type.__args__:
            if not (is_dataclass(union_item_type) or is_enum(union_item_type)):
                continue
            return cls._convert_field_value(union_item_type, field_value)
        return field_value

    @staticmethod
    def _convert_field_value_form_dict(field_cls: Any, field_dict_value: Dict[str, Any]) -> Any:
        if hasattr(field_cls, "from_dict"):
            field_value = getattr(field_cls, "from_dict")(**field_dict_value)
        else:
            field_value = field_cls(**field_dict_value)
        return field_value

    @classmethod
    def _convert_dataclass_field_type(cls, field_type: Any, field_value: Any) -> Any:
        if isinstance(field_value, dict):
            field_value = cls._convert_field_value_form_dict(field_type, field_value)
        field_value = cls._extend_convert_dataclass_field(field_type, field_value)
        return field_value

    @classmethod
    def _extend_convert_dataclass_field(cls, field_type: Any, field_value: Any) -> Any:
        return field_value

    @staticmethod
    def _convert_field_value_form_enum(field_cls: Type[Enum], field_value: Any) -> Any:
        has_form_value_func = hasattr(field_cls, "form_value")
        has_form_name_func = hasattr(field_cls, "form_name")
        if has_form_value_func:
            field_value = getattr(field_cls, "form_value")(field_value)
        elif has_form_name_func:
            field_value = getattr(field_cls, "form_name")(field_value)
        else:
            tmp_field_value = str(field_value).lower()
            for member_name, member in field_cls.__members__.items():
                if member_name.lower() == tmp_field_value:
                    return member
                if str(member.value).lower() == tmp_field_value:
                    return member

        return field_value

    @classmethod
    def _convert_field_value(cls, field_type: Any, field_value: Any) -> Any:
        if is_union(field_type):
            return cls._convert_union_field_type(field_type, field_value)

        if is_dataclass(field_type):
            try:
                field_value = cls._convert_dataclass_field_type(field_type, field_value)
            except Exception as exc:
                raise ConvertDataclassError(f"{exc} for '{field_type}'") from exc
        elif is_enum(field_type):
            field_value = cls._convert_field_value_form_enum(field_type, field_value)

        return field_value

    @classmethod
    def _from_dict(cls: Type[TDataclass], dacite_config: Optional[Config] = None, **data: Dict[str, Any]) -> 'TDataclass':
        if dacite_config is None:
            dacite_config = Config(
                type_hooks={datetime.datetime: lambda v: datetime_parse(v)},
                cast=[Decimal, float, int, Enum, uuid.UUID]
            )
        try:
            return from_dict(cls, data, config=dacite_config)
        except daciteExceptions.WrongTypeError as err:
            raise WrongTypeError(dataclass_cls=cls, field_type=err.field_type, value=err.value, field_path=err.field_path) from err

    @classmethod
    def from_dict(cls: Type[TDataclass], dacite_config: Optional[Config] = None, **data: Dict[str, Any]) -> 'TDataclass':
        values = {}
        for field_name, field in cls.__dataclass_fields__.items():  # type: ignore
            if field_name not in data:
                continue
            values[field_name] = cls._convert_field_value(field.type, data[field_name])

        return cls._from_dict(dacite_config=dacite_config, **values)

    def todict(self, dict_factory: Callable = dict, mode: DataclassAsdictModeEnum = DataclassAsdictModeEnum.DICT_ATTRIBUTE) -> Dict[str, Any]:
        if not is_dataclass(self):
            raise TypeError("todict() should be called on dataclass instances")
        return _asdict_inner(self, dict_factory, mode)

    def tojson(self) -> Dict[str, Any]:
        return pjson.loads(pjson.dumps(self))

    def __getattribute__(self, name: str) -> Any:
        v = super().__getattribute__(name)
        if isinstance(v, EmptyObj):
            raise AttributeError("'%s' object has no attribute '%s'" % (self.__class__.__name__, name))
        return v


@dataclass
class BaseModelDataclassMixin(BaseDataclassMixin):

    # use update db
    _changed_datas: Dict[str, Any] = field(repr=False, init=False)

    def __setattr__(self, name: str, value: Any) -> None:
        super().__setattr__(name, value)
        if not hasattr(self, "_changed_datas"):
            return
        if name != "_changed_datas":
            self._changed_datas[name] = value

    @property
    def changed_attributes(self) -> Dict[str, Any]:
        if not hasattr(self, "_changed_datas"):
            return {}
        return self._changed_datas

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._changed_datas = {}
        return False

    def __post_init__(self):
        self._changed_datas = {}
