import logging
import typing as t

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.core.exceptions import PermissionDenied as DJPermissionDenied
from django.http import Http404, HttpResponse
from django.http.response import HttpResponseBase
from rest_framework import exceptions as rfexceptions
from rest_framework.response import Response
from rest_framework.views import APIView, set_rollback

from mqer import pjson


UNKNOWNERROR_KEY = "UnknownError"
SUCCESS_CODE = "Ok"


class BaseAPIError(Exception):
    http_status_code = 500
    default_err_detail = "A server error occurred."
    default_err_code = "UnknownError"
    _headers = {}

    def __init__(self, err_detail: t.Optional[str] = None, err_code: t.Optional[str] = None) -> None:
        self.err_detail = err_detail if err_detail else self.default_err_detail
        self.err_code = err_code if err_code else self.default_err_code

    @property
    def headers(self) -> t.Dict[str, t.Any]:
        return self._headers

    @headers.setter
    def headers(self, headers: t.Dict[str, t.Any]):
        self._headers = headers


class NotFoundError(BaseAPIError):
    status_code = 404
    default_detail = 'Not found.'
    default_code = 'NotFoundError'


class PermissionDeniedError(BaseAPIError):
    http_status_code = 403
    default_err_detail = 'You do not have permission to perform this action.'
    default_err_code = 'PermissionDeniedError'


class IllegalParameterError(BaseAPIError):
    http_status_code = 400
    default_err_detail = 'Illegal request parameter.'
    default_err_code = 'IllegalParameterError'


def jsonify(body_data: t.Any, err_msg: str = "", err_code: str = SUCCESS_CODE, status_code: int = 200, encoder: t.Optional[t.Callable[[t.Any], t.Any]] = None, **kwargs) -> "HttpResponse":
    result_wrap = dict(err_msg=err_msg, err_code=err_code)
    if body_data:
        result_wrap["results"] = body_data

    body_content = pjson.dumps(result_wrap, encoder=encoder)
    kwargs.setdefault('content_type', 'application/json')
    return HttpResponse(content=body_content, status=status_code, **kwargs)


class BaseAPIView(APIView):
    def get_exception_handler(self):
        return exception_handler

    def success(self, body, encoder: t.Optional[t.Callable[[t.Any], t.Any]] = None, status_code: int = 200, **kwargs) -> "HttpResponse":
        return jsonify(body, encoder=encoder, status_code=status_code, **kwargs)

    def failed(self, err_code: t.Union[str, BaseAPIError], err_msg: str = "",  encoder: t.Optional[t.Callable[[t.Any], t.Any]] = None, status_code: t.Optional[int] = None, **kwargs) -> "HttpResponse":
        if isinstance(err_code, BaseAPIError):
            err_instance = err_code
            err_code = err_instance.err_code
            if not err_msg:
                err_msg = err_instance.err_detail
            if status_code is None:
                status_code = err_instance.http_status_code
        if status_code is None:
            status_code = 500
        return jsonify(None, err_code=err_code, err_msg=err_msg, encoder=encoder, status_code=status_code, **kwargs)

    def finalize_response(self, request, response, *args, **kwargs):
        if not isinstance(response, HttpResponseBase):
            try:
                return jsonify(response)
            except Exception as exc:
                response = self.handle_exception(exc)
        return super().finalize_response(request, response, *args, **kwargs)


def exception_handler(exc: Exception, context: t.Dict[str, t.Any]) -> 'Response':
    logger_name = getattr(settings, "EXCEPTION_LOGGER_NAME", "djapplication")
    logger = logging.getLogger(logger_name)

    headers = {}
    if getattr(exc, 'auth_header', None):
        headers['WWW-Authenticate'] = getattr(exc, 'auth_header')
    if getattr(exc, 'wait', None):
        headers['Retry-After'] = '%d' % getattr(exc, 'wait')

    err_details = None
    if isinstance(exc, rfexceptions.ValidationError):
        err_details = exc.get_full_details()
        reason_info = exc
        exc = IllegalParameterError()
        request = context.get("request", None)
        if request:
            reason_msg = f"Request url '{request.path}' request data validation failed, reason is '{str(reason_info)}'"
            logger.warning(reason_msg, extra={"err_details": err_details})
    elif isinstance(exc, (Http404, ObjectDoesNotExist)):
        exc = NotFoundError(str(exc))
    elif isinstance(exc, (DJPermissionDenied, rfexceptions.PermissionDenied)):
        exc = PermissionDeniedError(str(exc))

    result: t.Dict[str, t.Any] = {}
    if isinstance(exc, BaseAPIError):
        result = {'err_msg': exc.err_detail, 'err_code': exc.err_code}
        http_status_code = exc.http_status_code
    elif isinstance(exc, rfexceptions.APIException):
        source_err_detail = exc.detail
        if isinstance(source_err_detail, (list, dict)):
            err_code = exc.default_code.title().replace("_", "") if exc.default_code and exc.default_code != 'error' else UNKNOWNERROR_KEY
            err_msg = source_err_detail[0] if isinstance(source_err_detail, list) else tuple(source_err_detail.values())[0]
        else:
            err_code = source_err_detail.code.title().replace("_", "") if source_err_detail.code else UNKNOWNERROR_KEY
            err_msg = source_err_detail

        if not err_code.endswith("Error"):
            err_code = f"{err_code}Error"
        result = {'err_msg': err_msg, 'err_code': err_code}
        err_details = exc.get_full_details()
        http_status_code = exc.status_code
    else:
        result = {
            'err_msg': "Please check the application log, the reason may be '%s'." % exc,
            'err_code': UNKNOWNERROR_KEY
        }
        http_status_code = 500
        logger.exception("Exception in application business, the reason may be '%s'", exc)

    set_rollback()
    if err_details:
        result["err_details"] = err_details
    return Response(result, status=http_status_code, headers=headers)
