from enum import Enum, EnumMeta
from typing import Any, List, Optional, Tuple, Type, TypeVar

TEnum = TypeVar("TEnum", bound="Enum")


def is_enum(value: Any) -> bool:
    if hasattr(value, "__members__"):
        return True

    try:
        return issubclass(value.__class__, (EnumMeta, Enum))
    except TypeError:
        return False


class BaseEnumMixin(Enum):
    @classmethod
    def options(cls) -> List[Tuple[str, Any]]:
        return [(name, member.value) for name, member in cls.__members__.items()]

    @classmethod
    def form_value(cls: Type[TEnum], member_value: Any, raise_exception: bool = True) -> Optional['TEnum']:
        _member_value = member_value.value if is_enum(member_value) else member_value
        _member_value = str(_member_value).upper()
        for member in cls.__members__.values():
            if str(member.value).upper() == _member_value:
                return member
        if not raise_exception:
            return None
        raise ValueError(f"Record not matched of `{cls}`. member value is `{member_value}`")

    @classmethod
    def form_name(cls: Type[TEnum], member_name: str, raise_exception: bool = True) -> Optional['TEnum']:
        member_name = member_name.upper()
        for name, member in cls.__members__.items():
            if name.upper() == member_name:
                return member
        if not raise_exception:
            return None
        raise ValueError(f"Record not matched of `{cls}`. member name is `{member_name}`")

    @classmethod
    def values(cls) -> List[Any]:
        return [member.value for _, member in cls.__members__.items()]

    @classmethod
    def names(cls) -> List[str]:
        return [name for name, _ in cls.__members__.items()]
