import typing as t
from functools import partial
from types import TracebackType

from mqer import pjson
from mqer.local import LocalProxy, LocalStack

_sentinel = object()
_log_ctx_stack = LocalStack()


def _ctx_lookup(name: str) -> t.Any:
    top = _log_ctx_stack.top
    if top is None:
        raise RuntimeError(f"Attempt to access {name} outside of a relevant context")
    return getattr(top, name)


logextra: 'LogExtraDataclass' = LocalProxy(partial(_ctx_lookup, "logextra"))


class LogExtraDataclass:
    def __getattr__(self, name: str) -> t.Any:
        try:
            return self.__dict__[name]
        except KeyError:
            raise AttributeError(name) from None

    def __setattr__(self, name: str, value: t.Any) -> None:
        self.__dict__[name] = value

    def __delattr__(self, name: str) -> None:
        try:
            del self.__dict__[name]
        except KeyError:
            raise AttributeError(name) from None

    def __contains__(self, item: str) -> bool:
        return item in self.__dict__

    def __iter__(self) -> t.Iterator[str]:
        return iter(self.__dict__)

    def get(self, name: str, default: t.Optional[t.Any] = None) -> t.Any:
        return self.__dict__.get(name, default)

    def pop(self, name: str, default: t.Any = _sentinel) -> t.Any:
        if default is _sentinel:
            return self.__dict__.pop(name)
        else:
            return self.__dict__.pop(name, default)

    def setdefault(self, name: str, default: t.Any = None) -> t.Any:
        return self.__dict__.setdefault(name, default)

    def todict(self) -> t.Dict[str, t.Any]:
        return self.__dict__


class LogExtraContext:
    def __init__(self) -> None:
        self.logextra = LogExtraDataclass()

    def push(self) -> None:
        _log_ctx_stack.push(self)

    def pop(self, exc: t.Optional[BaseException] = _sentinel) -> None:
        rv = _log_ctx_stack.pop()
        assert rv is self, f"Popped wrong logextra context.  ({rv!r} instead of {self!r})"

    def __enter__(self) -> "LogExtraContext":
        self.push()
        return self

    def __exit__(self, exc_type: t.Optional[type], exc_value: t.Optional[BaseException], tb: t.Optional[TracebackType]) -> None:
        self.pop(exc_value)


class StructuredLogMessage:
    msg: str
    options: t.Dict[str, t.Any]

    # from https://docs.python.org/3.8/howto/logging-cookbook.html#implementing-structured-logging
    def __init__(self, msg: str, **options) -> None:
        self.msg = msg
        self.options = options

    def __str__(self) -> str:
        return f"{self.msg} :: {', '.join(k+ ': ' + pjson.dumps(v) for k,v in self.options.items())}"


def has_logextra_context() -> bool:
    return _log_ctx_stack.top is not None
