import logging
import typing as t
from datetime import datetime

from pythonjsonlogger import jsonlogger

from mqer import pjson
from mqer.jsonlogger import StructuredLogMessage, has_logextra_context, logextra


def json_serializer(o: t.Any, *args, **kwargs) -> str:
    json_indent = kwargs.pop("indent", None)
    indent = True if json_indent is not None else False
    dumps_options: t.Dict[str, t.Any] = dict(indent=indent)

    if kwargs.get("default") and callable(kwargs["default"]):
        dumps_options["encoder"] = kwargs["default"]

    if kwargs.get("json_option") is not None:
        dumps_options["option"] = int(kwargs["json_option"])

    return pjson.dumps(o, **dumps_options)


class JsonFormatter(jsonlogger.JsonFormatter):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("timestamp", True)
        self.json_option = kwargs.pop("json_option", None)
        self.is_native_json_serializer = kwargs.pop("is_native_json_serializer", False)
        if not self.is_native_json_serializer and "json_serializer" not in kwargs:
            kwargs["json_serializer"] = json_serializer
        super().__init__(*args, **kwargs)

    def jsonify_log_record(self, log_record):
        """Returns a json string of the log record."""
        if self.is_native_json_serializer:
            return super().jsonify_log_record(log_record)

        return self.json_serializer(
            log_record,
            default=self.json_default,
            cls=self.json_encoder,
            indent=self.json_indent,
            ensure_ascii=self.json_ensure_ascii,
            json_option=self.json_option
        )

    def format(self, rec: logging.LogRecord) -> str:
        if isinstance(rec.msg, StructuredLogMessage):
            ans = {"level": rec.levelname, "message": rec.msg.msg}
            for k, v in rec.msg.options.items():
                if k not in ans:
                    ans[k] = v
            rec.msg = ans
        return super().format(rec)

    def add_fields(self, log_record: t.Dict[str, t.Any], record: logging.LogRecord, message_dict: t.Dict[str, t.Any]):
        super().add_fields(log_record, record, message_dict)
        if not log_record.get("timestamp"):
            # this doesn't use record.created, so it is slightly off
            log_record['timestamp'] = datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%fZ')

        # add logextra context value
        if has_logextra_context():
            for k, v in logextra.todict().items():
                log_record[k] = v
