import decimal
import functools
import datetime
from dataclasses import is_dataclass, asdict
from typing import Any, Callable, Optional, Union

import orjson

from mqer.types import EmptyObj


OPT_APPEND_NEWLINE = orjson.OPT_APPEND_NEWLINE
OPT_INDENT_2 = orjson.OPT_INDENT_2
OPT_NAIVE_UTC = orjson.OPT_NAIVE_UTC
OPT_NON_STR_KEYS = orjson.OPT_NON_STR_KEYS
OPT_OMIT_MICROSECONDS = orjson.OPT_OMIT_MICROSECONDS
OPT_PASSTHROUGH_DATACLASS = orjson.OPT_PASSTHROUGH_DATACLASS
OPT_PASSTHROUGH_DATETIME = orjson.OPT_PASSTHROUGH_DATETIME
OPT_PASSTHROUGH_SUBCLASS = orjson.OPT_PASSTHROUGH_SUBCLASS
OPT_SERIALIZE_DATACLASS = orjson.OPT_SERIALIZE_DATACLASS
OPT_SERIALIZE_NUMPY = orjson.OPT_SERIALIZE_NUMPY
OPT_SERIALIZE_UUID = orjson.OPT_SERIALIZE_UUID
OPT_SORT_KEYS = orjson.OPT_SORT_KEYS
OPT_STRICT_INTEGER = orjson.OPT_STRICT_INTEGER
OPT_UTC_Z = orjson.OPT_UTC_Z

DEFAULT_OPTIONS: int = OPT_UTC_Z | OPT_NAIVE_UTC | OPT_PASSTHROUGH_DATACLASS | OPT_NON_STR_KEYS
SIMPLE_OPTIONS: int = OPT_UTC_Z | OPT_NAIVE_UTC

class DefaultJSONEncoder:
    @staticmethod
    def encoder(obj: Any):
        if isinstance(obj, decimal.Decimal):
            # Serializers will coerce decimals to strings by default.
            return float(obj)
        if hasattr(obj, 'todict') and callable(getattr(obj, "todict", None)):
            # Method of transforming dict structure from custom dataclass
            return obj.todict()
        if hasattr(obj, 'tojson') and callable(getattr(obj, "tojson", None)):
            return obj.tojson()
        if is_dataclass(obj):
            return asdict(obj)
        if isinstance(obj, datetime.timedelta):
            return str(obj.total_seconds())
        if not isinstance(obj, dict) and hasattr(obj, '__getitem__'):
            return dict(obj)
        if not isinstance(obj, (list, tuple)) and hasattr(obj, '__iter__'):
            return tuple(item for item in obj)

        if isinstance(obj, EmptyObj):
            return None

        raise TypeError


@functools.wraps(orjson.dumps)
def dumps(obj: Any, encoder: Optional[Callable[[Any], Any]] = None, indent: bool = False, option: int = DEFAULT_OPTIONS, **kwargs) -> str:
    if encoder is None:
        encoder = DefaultJSONEncoder.encoder

    if indent:
        option = option | OPT_INDENT_2

    kwargs["option"] = option

    return orjson.dumps(obj, default=encoder, **kwargs).decode("utf-8")


@functools.wraps(orjson.loads)
def loads(obj: Union[bytes, bytearray, memoryview, str]) -> Any:
    return orjson.loads(obj)
