import logging
from collections import defaultdict
from concurrent import futures
from concurrent.futures.process import BrokenProcessPool
from dataclasses import dataclass, field
from threading import RLock
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from mqer.logger import logger as mqer_logger

PoolExecutorVar = Union[futures.ProcessPoolExecutor, futures.ThreadPoolExecutor]
LoggerVar = Union[logging.Logger, logging.LoggerAdapter, str]


@dataclass(frozen=True)
class Job:
    id: str
    func: Callable[..., Any]
    func_args: Tuple[Any, ...] = field(default_factory=tuple)
    func_kwargs: Dict[Any, Any] = field(default_factory=dict)
    max_instances: int = 1
    after_run_callbacks: List[Callable[["Job", futures.Future], Any]] = field(default_factory=list)
    logger: Optional[LoggerVar] = None


class MaxInstancesReachedError(Exception):
    def __init__(self, job):
        super(MaxInstancesReachedError, self).__init__(f'Job "{job.id}" has already reached its maximum number of instances ({job.max_instances}).')


def to_logger_instance(logger_or_name: LoggerVar) -> Union[logging.Logger, logging.LoggerAdapter]:
    if isinstance(logger_or_name, str):
        _logger = logging.getLogger(logger_or_name)
    else:
        _logger = logger_or_name
    return _logger


def _run_job(job: Job, logger: LoggerVar) -> Any:
    """
    Called by executors to run the job.
    """
    job_logger = to_logger_instance(job.logger if job.logger is not None else logger)
    job_logger.info('Running job "%s"', job)
    try:
        result = job.func(*job.func_args, **job.func_kwargs)
        job_logger.info('Job "%s" executed successfully', job)
        return result
    except BaseException as exc:
        job_logger.exception('Job "%s" raised an exception: %s', job, exc)
        raise exc


@dataclass
class _JobFutureMapper:
    _futures: Dict[int, futures.Future] = field(default_factory=dict)

    def add(self, ft: 'futures.Future'):
        self._futures[id(ft)] = ft

    def remove(self, ft: 'futures.Future'):
        self._futures.pop(id(ft), None)

    def cancel(self):
        for ft in tuple(self._futures.keys()):
            self._futures[ft].cancel()

    @property
    def instance_count(self) -> int:
        return len(self._futures)


class BasePoolExecutor:
    def __init__(self, pool: PoolExecutorVar, logger: Optional[LoggerVar] = None, run_job_func: Optional[Callable[["Job", LoggerVar], Any]] = None) -> None:
        self._pool = pool
        self._instances = defaultdict(_JobFutureMapper)
        self._logger = to_logger_instance(logger if logger is not None else mqer_logger)
        self._lock = RLock()
        self.run_job_func = run_job_func if run_job_func is not None else _run_job

    def submit_job(self, job: Job) -> "futures.Future":
        """
        Submits job for execution.
        :param Job job: job to execute
        :raises MaxInstancesReachedError: if the maximum number of
            allowed instances for this job has been reached
        """
        with self._lock:
            if self._instances[job.id].instance_count >= job.max_instances:
                raise MaxInstancesReachedError(job)

            ft = self._do_submit_job(job)
            self._instances[job.id].add(ft)
            return ft

    def _do_submit_job(self, job: Job) -> "futures.Future":
        def callback(ft: futures.Future):
            if hasattr(ft, 'exception_info'):
                exc, tb = getattr(ft, "exception_info")()
            else:
                exc, tb = (ft.exception(), getattr(ft.exception(), '__traceback__', None)) if not ft.cancelled() else (None, None)

            job_logger = to_logger_instance(job.logger if job.logger is not None else self._logger)

            if exc:
                exc_info = (exc.__class__, exc, tb)
                job_logger.error('Error running job %s', job, exc_info=exc_info)

            job_id = job.id
            with self._lock:
                self._instances[job_id].remove(ft)
                if self._instances[job_id].instance_count == 0:
                    del self._instances[job_id]
            try:
                for func in job.after_run_callbacks:
                    func(job, ft)
            except Exception as exc:
                job_logger.exception('Job "%s" callback run after func raised an exception: %s', job, exc)

        try:
            f = self._pool.submit(self.run_job_func, job, self._logger)
        except BrokenProcessPool:
            self._logger.warning('Process pool is broken; replacing pool with a fresh instance.')
            self._pool = self._pool.__class__(getattr(self._pool, "_max_workers", 1))
            f = self._pool.submit(self.run_job_func, job, self._logger)

        f.add_done_callback(callback)
        return f

    def cancel(self, job_id: str) -> bool:
        self._logger.warning("Cancel job '%s' futures" % job_id)
        with self._lock:
            if job_id not in self._instances:
                return False
            _jobftmapper = self._instances[job_id]
            _jobftmapper.cancel()
        return True

    def shutdown(self, wait=True):
        self._logger.warning("Shutting down pool executor, wait=%s", wait)
        with self._lock:
            for job_id in tuple(self._instances.keys()):
                self.cancel(job_id)
        self._pool.shutdown(wait)


@dataclass(frozen=True)
class PoolInitializer:
    init_func: Optional[Callable[..., Any]] = None
    init_args: Tuple[Any, ...] = field(default_factory=tuple)
    run_job_func: Optional[Callable[["Job", LoggerVar], Any]] = None


class ProcessPoolExecutor(BasePoolExecutor):
    def __init__(self, max_workers: int, logger: Optional[LoggerVar] = None, initializer: Optional[PoolInitializer] = None) -> None:
        init_func, init_args, run_job_func = None, (), None
        if initializer is not None:
            init_func, init_args, run_job_func = initializer.init_func, initializer.init_args, initializer.run_job_func
        pool = futures.ProcessPoolExecutor(max_workers=max_workers, initializer=init_func, initargs=init_args)
        super(ProcessPoolExecutor, self).__init__(pool, logger=logger, run_job_func=run_job_func)


class ThreadPoolExecutor(BasePoolExecutor):
    def __init__(self, max_workers: int, logger: Optional[LoggerVar] = None, initializer: Optional[PoolInitializer] = None) -> None:
        init_func, init_args, run_job_func = None, (), None
        if initializer is not None:
            init_func, init_args, run_job_func = initializer.init_func, initializer.init_args, initializer.run_job_func
        pool = futures.ThreadPoolExecutor(max_workers=max_workers, initializer=init_func, initargs=init_args)
        super(ThreadPoolExecutor, self).__init__(pool, logger=logger, run_job_func=run_job_func)
