import signal
from typing import Any, Callable, Dict, List, Optional, Union

from apscheduler import events
from apscheduler.executors.pool import ProcessPoolExecutor, ThreadPoolExecutor
from apscheduler.job import Job as apschedulerJob
from apscheduler.schedulers.blocking import BlockingScheduler
from apscheduler.triggers.cron import CronTrigger

from mqer.logger import logger
from mqer.types import CallbackFunc
from mqer.utils import import_obj_string

Job = apschedulerJob


class SecondCronTrigger(CronTrigger):
    @classmethod
    def from_crontab(cls, expr: str, timezone=None):
        """
        Create a :class:`~SecondCronTrigger` from a standard crontab expression.

        :param expr: second, minute, hour, day of month, month, day of week
        :param datetime.tzinfo|str timezone: time zone to use for the date/time calculations (defaults to scheduler timezone)
        :return: a :class:`~SecondCronTrigger` instance
        """
        values = expr.split()
        if len(values) != 6:
            raise ValueError('Wrong number of fields; got {}, expected 6'.format(len(values)))

        return cls(second=values[0], minute=values[1], hour=values[2], day=values[3], month=values[4], day_of_week=values[5], timezone=timezone)


class Timer:
    def __init__(self, threads: int = 1, process: int = 1, timezone: str = "UTC", listener_funcs: Optional[List[Callable[..., Any]]] = None,
                 before_run_funcs: Optional[List[CallbackFunc]] = None, after_shutdown_funcs: Optional[List[CallbackFunc]] = None):
        self.threads = threads
        self.process = process
        self.timezone = timezone
        self.listener_funcs = listener_funcs if listener_funcs is not None else []
        self.before_run_funcs = before_run_funcs if before_run_funcs is not None else []
        self.after_shutdown_funcs = after_shutdown_funcs if after_shutdown_funcs is not None else []
        executors = {'default': ThreadPoolExecutor(self.threads), 'processpool': ProcessPoolExecutor(self.process)}
        self.scheduler = BlockingScheduler(executors=executors, timezone=self.timezone)

    def register_listener_func(self, func: Callable[..., Any]):
        self.listener_funcs.append(func)

    def register_before_run_func(self, func: CallbackFunc):
        self.before_run_funcs.append(func)

    def register_after_run_func(self, func: CallbackFunc):
        self.after_shutdown_funcs.append(func)

    def job(self, job_id: str, cron_expr: str, max_instances: int = 1, args: Optional[List[Any]] = None, kwargs: Optional[Dict[Any, Any]] = None,
            coalesce: bool = True, replace_existing: bool = True) -> Callable[..., Any]:
        def _wrapper(func):
            self.add_job(
                job_id=job_id,
                func=func,
                args=args,
                kwargs=kwargs,
                cron_expr=cron_expr,
                max_instances=max_instances,
                coalesce=coalesce,
                replace_existing=replace_existing
            )
        return _wrapper

    def add_job(self, job_id: str, func: Union[Callable[..., Any], str], cron_expr: str, max_instances: int = 1, args: Optional[List[Any]] = None, kwargs: Optional[Dict[Any, Any]] = None,
                coalesce: bool = True, replace_existing: bool = True) -> "Job":
        job = self.scheduler.add_job(
            id=job_id,
            func=func,
            args=args,
            kwargs=kwargs,
            trigger=SecondCronTrigger.from_crontab(expr=cron_expr),
            max_instances=max_instances,
            coalesce=coalesce,
            replace_existing=replace_existing
        )
        return job

    def setup_from_conf(self, job_confs: Dict[str, Any]):
        for job_id, job_conf in job_confs.items():
            self.add_job(
                job_id=job_id,
                func=import_obj_string(job_conf["task"]),
                cron_expr=job_conf["cron_expr"],
                max_instances=job_conf.get("max_instances", 1),
                coalesce=job_conf.get("coalesce", True),
                replace_existing=job_conf.get("replace_existing", True),
                args=job_conf.get("args", None),
                kwargs=job_conf.get("kwargs", None)
            )

    def start(self):
        mask = events.EVENT_JOB_ERROR | events.EVENT_JOB_EXECUTED
        for func in self.listener_funcs:
            self.scheduler.add_listener(func, mask)

        for callback in self.before_run_funcs:
            callback.func(*callback.args, **callback.kwargs)

        signal.signal(signal.SIGTERM, self._shutdown)
        try:
            logger.debug("Starting mqer timer, job qty is %s", len(self.scheduler.get_jobs()))
            self.scheduler.start()
        except KeyboardInterrupt:
            self._shutdown()

    def _shutdown(self, *args, **kwargs):
        logger.debug("Shutdown mqer timer worker.")
        self.scheduler.shutdown()

        for callback in self.after_shutdown_funcs:
            try:
                callback.func(*callback.args, **callback.kwargs)
            except Exception as exc:
                logger.exception("Callback function `%s` handler raise exception after timer shutdown, reason may be: %s", callback, exc)
                continue

    @classmethod
    def from_conf(cls, timer_conf: Dict[str, Any]) -> "Timer":
        threads = timer_conf.get("threads", 1)
        process = timer_conf.get("process", 1)
        timezone = timer_conf.get("timezone", "UTC")
        listener_funcs: List[Callable[..., Any]] = [import_obj_string(f) if isinstance(f, str) else f for f in timer_conf.get("listener_funcs", [])]
        before_run_funcs: List[CallbackFunc] = [cf if isinstance(cf, CallbackFunc) else CallbackFunc.from_dict(cf)
                                                for cf in timer_conf.get("before_run_funcs", [])]
        after_shutdown_funcs: List[CallbackFunc] = [cf if isinstance(cf, CallbackFunc) else CallbackFunc.from_dict(cf)
                                                    for cf in timer_conf.get("after_shutdown_funcs", [])]
        jobs = timer_conf.get("jobs", [])

        timer_instance = cls(
            threads=threads,
            process=process,
            timezone=timezone,
            listener_funcs=listener_funcs,
            before_run_funcs=before_run_funcs,
            after_shutdown_funcs=after_shutdown_funcs
        )
        timer_instance.setup_job_from_conf(jobs)
        return timer_instance

    def setup_job_from_conf(self, job_confs: List[Dict[str, Any]]):
        for job_conf in job_confs:
            self.add_job(
                job_id=job_conf.get("job_id", ''),
                func=import_obj_string(job_conf["task"]) if isinstance(job_conf["task"], str) else job_conf["task"],
                cron_expr=job_conf["cron_expr"],
                max_instances=job_conf.get("max_instances", 1),
                coalesce=job_conf.get("coalesce", True),
                replace_existing=job_conf.get("replace_existing", True),
                args=job_conf.get("args", None),
                kwargs=job_conf.get("kwargs", None)
            )
