import warnings
from typing import Any, Dict, List, Optional, Union

from kombu import Connection, Exchange, Queue

from mqer.types import JsonMsgBody
from mqer.utils import make_msg_uuid
from mqer.worker.message import (
    MSG_TRACKID_KEY,
    MSGCLS_PATH_HEADER_KEY,
    BaseMessage
)


class RemovedInMQER104Warning(DeprecationWarning):
    pass


class BaseMQClient:
    def __init__(self, conn_url: str, confirm_publish: bool = True) -> None:
        # amqp://guest:guest@localhost:5672//
        self.conn_url = conn_url
        self.confirm_publish = confirm_publish

    @property
    def connection(self) -> 'Connection':
        return Connection(self.conn_url, transport_options={"confirm_publish": self.confirm_publish})

    def _push_msg(self, msg_body: JsonMsgBody, exchange: 'Exchange', routing_key: str, queues: List['Queue'],
                  headers: Optional[Dict[str, Any]] = None, retry: bool = False) -> str:
        if headers is None:
            headers = {}

        if MSG_TRACKID_KEY not in headers:
            headers[MSG_TRACKID_KEY] = make_msg_uuid()

        with self.connection as conn:
            channel = conn.channel()
            producer = conn.Producer(channel=channel, serializer='json', exchange=exchange, routing_key=routing_key)
            # declares exchange, queue and binds before sending message
            producer.publish(body=msg_body, declare=queues, headers=headers, retry=retry)

        return headers[MSG_TRACKID_KEY]

    def _publish(self, exchange: 'Exchange', queue: 'Queue', message: 'BaseMessage') -> str:
        warnings.warn(
            '_publish() is deprecated in favor of publish().',
            RemovedInMQER104Warning, stacklevel=2,
        )
        return self.publish(exchange, queue, message)

    def _broadcast_publish(self, exchange: 'Exchange', queues: List['Queue'], message: 'BaseMessage') -> str:
        warnings.warn(
            '_broadcast_publish() is deprecated in favor of broadcast_publish().',
            RemovedInMQER104Warning, stacklevel=2,
        )
        return self.broadcast_publish(exchange, queues, message)

    def broadcast_publish(self, exchange: 'Exchange', queues: List['Queue'], message: Union['BaseMessage', 'JsonMsgBody'],
                          retry: bool = False) -> str:
        if isinstance(message, BaseMessage):
            msg_headers = {MSGCLS_PATH_HEADER_KEY: message.msgcls_path}
            msg_body = message.todict()
        else:
            msg_headers = {}
            msg_body = message

        return self._push_msg(
            msg_body=msg_body,
            exchange=exchange,
            routing_key="",
            queues=queues,
            headers=msg_headers,
            retry=retry
        )

    def publish(self, exchange: 'Exchange', queue: 'Queue', message: Union['BaseMessage', 'JsonMsgBody'],
                retry: bool = False) -> str:
        if isinstance(message, BaseMessage):
            msg_headers = {MSGCLS_PATH_HEADER_KEY: message.msgcls_path}
            msg_body = message.todict()
        else:
            msg_headers = {}
            msg_body = message

        return self._push_msg(
            msg_body=msg_body,
            exchange=exchange,
            routing_key=queue.routing_key,
            queues=[queue],
            headers=msg_headers,
            retry=retry
        )

    @staticmethod
    def make_exchange(name: str, exchange_type: str, durable: bool = True, **kwargs) -> 'Exchange':
        return Exchange(name=name, type=exchange_type, durable=durable, **kwargs)

    @staticmethod
    def make_queue(name: str, exchange: 'Exchange', routing_key: str, durable: bool = True, auto_delete: bool = False, **kwargs) -> 'Queue':
        return Queue(name=name, exchange=exchange, routing_key=routing_key, durable=durable, auto_delete=auto_delete, **kwargs)
