import json
import signal
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, List, Optional, Tuple

from dacite import from_dict
from kombu import Connection, Message, Queue
from kombu.mixins import ConsumerMixin

from mqer.logger import logger
from mqer.worker.message import DISCARDED_HEADER_KEY, MSGCLS_PATH_HEADER_KEY, BaseMessage
from mqer.types import CallbackFunc
from mqer.utils import import_obj_string

DEFAULT_ORIGIN_MSG_KEY = "__pillardefaultmsgkey__"


@dataclass
class MsgSerializer:
    originmsgkey: str
    destmsgcls: str

    def decode(self, msg_key: str) -> Tuple[bool, Any]:
        if self.originmsgkey == DEFAULT_ORIGIN_MSG_KEY:
            return True, import_obj_string(msg_key)

        if self.originmsgkey == msg_key:
            return True, import_obj_string(self.destmsgcls)

        return False, None


@dataclass
class Receiver:
    func: Callable[..., Any]
    serializer: MsgSerializer


@dataclass
class QueueReceiver:
    exchange: str
    exchange_type: str
    queue: str
    routing_key: str
    durable: bool
    auto_delete: bool
    receivers: List[Receiver] = field(default_factory=list)

    @classmethod
    def from_dict(cls, **data: Dict[str, Any]) -> 'QueueReceiver':
        return from_dict(cls, data)


@dataclass
class consumer:
    queues: List['Queue']
    receivers: List['Receiver']


class ConsumWorker(ConsumerMixin):
    def __init__(self, connection: 'Connection', consumers: List['consumer']) -> None:
        self.connection = connection
        self.consumers = consumers

    def get_consumers(self, Consumer, channel):
        return [Consumer(queues=c.queues, callbacks=[self.create_msg_handler(c.receivers)], prefetch_count=1) for c in self.consumers]

    def create_msg_handler(self, receivers: List['Receiver']):
        def on_msg_received(body, message: 'Message'):
            message_headers = message.headers
            messageclass_path = message_headers.get(MSGCLS_PATH_HEADER_KEY)
            discarded = message_headers.get(DISCARDED_HEADER_KEY, False)
            if discarded:
                message.ack()
                return

            if not messageclass_path:
                receivers[0].func(body)
                message.ack()
                return

            for receiver in receivers:
                is_convert, msg_cls = receiver.serializer.decode(messageclass_path)
                if not is_convert:
                    continue

                msg_cls: 'BaseMessage' = msg_cls
                if isinstance(body, str):
                    body = json.loads(body)
                msg_obj = msg_cls.from_dict(**body)
                receiver.func(msg_obj)
                message.ack()
                break
        return on_msg_received


class Worker:
    def __init__(self, conn_url: str, before_run_funcs: Optional[List[CallbackFunc]] = None, after_shutdown_funcs: Optional[List[CallbackFunc]] = None):
        self.conn_url = conn_url
        self.before_run_funcs = before_run_funcs if before_run_funcs is not None else []
        self.after_shutdown_funcs = after_shutdown_funcs if after_shutdown_funcs is not None else []
        self._func_receivers: Dict[str, consumer] = {}

    def register_before_run_func(self, func: CallbackFunc):
        self.before_run_funcs.append(func)

    def register_after_run_func(self, func: CallbackFunc):
        self.after_shutdown_funcs.append(func)

    def receiver(self, queue_name: str, routing_key: str, durable: bool = True, auto_delete: bool = False, msg_serializer: Optional[MsgSerializer] = None) -> Callable[..., Any]:
        def _wrapper(func):
            self.add_receiver(queue_name, routing_key, func, durable, auto_delete, msg_serializer)
        return _wrapper

    def add_receiver(self, queue_name: str, routing_key: str, func: Callable[..., Any], durable: bool = True, auto_delete: bool = False, msg_serializer: Optional[MsgSerializer] = None):
        if msg_serializer is None:
            originmsgkey, destmsgcls = DEFAULT_ORIGIN_MSG_KEY, DEFAULT_ORIGIN_MSG_KEY
        else:
            originmsgkey, destmsgcls = msg_serializer.originmsgkey, msg_serializer.destmsgcls

        self._make_receiver(QueueReceiver(
            exchange="",
            exchange_type="",
            queue=queue_name,
            routing_key=routing_key,
            durable=durable,
            auto_delete=auto_delete,
            receivers=[Receiver(func=func, serializer=MsgSerializer(originmsgkey=originmsgkey, destmsgcls=destmsgcls))]
        ))

    def _make_receiver(self, queue_receiver: QueueReceiver):
        queue_name = queue_receiver.queue
        if queue_name not in self._func_receivers:
            queue = Queue(
                name=queue_name,
                routing_key=queue_receiver.routing_key,
                durable=queue_receiver.durable,
                auto_delete=queue_receiver.auto_delete
            )
            self._func_receivers[queue_name] = consumer(queues=[queue], receivers=[])

        self._func_receivers[queue_name].receivers.extend(queue_receiver.receivers)

    def setup_from_conf(self, queue_receiver_confs: Dict[str, Any]):
        for queue_receiver_conf in queue_receiver_confs.values():
            receivers: List['Receiver'] = []
            for receiver_conf in queue_receiver_conf.get("receivers", []):
                receivers.append(Receiver(
                    func=import_obj_string(receiver_conf["func"]),
                    serializer=MsgSerializer(originmsgkey=receiver_conf["originmsgkey"], destmsgcls=receiver_conf["destmsgcls"])
                ))
            if not receivers:
                logger.warn("No found queue receiver for queue `%s`.", queue_receiver_conf["queue"])
                continue

            self._make_receiver(QueueReceiver(
                queue=queue_receiver_conf["queue"],
                routing_key=queue_receiver_conf["routing_key"],
                durable=queue_receiver_conf.get("durable", True),
                auto_delete=queue_receiver_conf.get("auto_delete", False),
                exchange=queue_receiver_conf.get("exchange", "amq.direct"),
                exchange_type=queue_receiver_conf.get("exchange_type", "direct"),
                receivers=receivers
            ))

    def start(self):
        for callback in self.before_run_funcs:
            callback.func(*callback.args, **callback.kwargs)

        signal.signal(signal.SIGTERM, self._shutdown)
        with Connection(self.conn_url) as c:
            try:
                worker = ConsumWorker(c, list(self._func_receivers.values()))
                logger.info("Starting mqer worker, conn_url is %s", self.conn_url)
                worker.run()
            except KeyboardInterrupt:
                self._shutdown()

    def _shutdown(self, *args, **kwargs):
        logger.debug("Shutdown mqer worker.")
        for callback in self.after_shutdown_funcs:
            try:
                callback.func(*callback.args, **callback.kwargs)
            except Exception as exc:
                logger.exception("Callback function `%s` handler raise exception after worker shutdown, reason may be: %s", callback, exc)
                continue
